/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * operators.hpp
 *
 *  Created on: 2010/11/18
 *      Author: HARK Development Team
 */

#ifndef DIMPRO_OPERATORS_HPP_
#define DIMPRO_OPERATORS_HPP_

/**
 * @file types.hpp
 * 単位保証に使う演算を集めたファイル.
 * 複数の型に共通する演算や複数の型を用いなければ定義できない演算など
 */

#include <limits>
#include "types.hpp"

namespace NSDimensionPromise
{

/**
 * テンプレートメタプログラミング用のif関数.
 * condがtrueならThenを返す
 */
template <bool cond, typename Then, typename Else>
struct IfStruct
{
    typedef Then type;
};

/**
 * テンプレートメタプログラミング用のif関数のfalse特殊化.
 */
template <typename Then, typename Else>
struct IfStruct<false, Then, Else>
{
    typedef Else type;
};

/**
 * 2つの型のうち、表現可能な桁数が大きな型を返す.
 * T1が表現可能な桁数とT2が表現可能な桁数を比較し、
 * T1の表現可能な桁数の方が大きいか等しければ、T1を返す
 */
template <typename T1, typename T2>
struct BiggerDigitsType
{
    /// @cond
    /**
     * T1とT2のうち表現可能な桁数が大きい方をtypeとする.
     */
    typedef typename IfStruct<(std::numeric_limits<T1>::digits10 >= std::numeric_limits<T2>::digits10), T1, T2>::type type;
    /// @endcond
};

/**
 * スカラーの乗算.
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename L, typename R,
         int Lm, int Lkg, int Ls, int LA, int LK, int Lmol, int Lcd,
         int Rm, int Rkg, int Rs, int RA, int RK, int Rmol, int Rcd
         >
inline
const ScalarNumber<typename BiggerDigitsType<L, R>::type,
      Lm+Rm, Lkg+Rkg, Ls+Rs, LA+RA, LK+RK, Lmol+Rmol, Lcd+Rcd>
      operator *(
          const ScalarNumber<L, Lm, Lkg, Ls, LA, LK, Lmol, Lcd>& lhs,
          const ScalarNumber<R, Rm, Rkg, Rs, RA, RK, Rmol, Rcd>& rhs
      )
{
    return ScalarNumber<typename BiggerDigitsType<L, R>::type,
           Lm+Rm, Lkg+Rkg, Ls+Rs, LA+RA, LK+RK, Lmol+Rmol, Lcd+Rcd>(
               lhs.X() * rhs.X());
}

#if 0
/**
 * スカラーの乗算.
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename T,
int m, int kg, int s, int A, int K, int mol, int cd
>
inline
const ScalarNumber<T, m, kg, s, A, K, mol, cd>
operator *(
    const ScalarNumber<T, m, kg, s, A, K, mol, cd>& lhs,
    const T& rhs
)
{
    return ScalarNumber<T, m, kg, s, A, K, mol, cd>(
               lhs.X() * rhs);
}
#else
/**
 * スカラーの乗算.
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename T,
int m, int kg, int s, int A, int K, int mol, int cd
>
inline
const ScalarNumber<T, m, kg, s, A, K, mol, cd>
operator *(
    const ScalarNumber<T, m, kg, s, A, K, mol, cd>& lhs,
    double rhs // const T&ではなく、doubleにすることで、floatやintも暗黙の型変換によって引数に取れるようになる。
)
{
    return ScalarNumber<T, m, kg, s, A, K, mol, cd>(
               lhs.X() * rhs);
}

#endif

/**
 * スカラーの乗算.
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename T,
         int m, int kg, int s, int A, int K, int mol, int cd
         >
inline
const ScalarNumber<T, m, kg, s, A, K, mol, cd>
operator *(
    const T& lhs,
    const ScalarNumber<T, m, kg, s, A, K, mol, cd>& rhs
)
{
    return ScalarNumber<T, m, kg, s, A, K, mol, cd>(
               lhs * rhs.X());
}

/**
 * スカラーの除算.
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename L, typename R,
         int Lm, int Lkg, int Ls, int LA, int LK, int Lmol, int Lcd,
         int Rm, int Rkg, int Rs, int RA, int RK, int Rmol, int Rcd
         >
inline
const ScalarNumber<typename BiggerDigitsType<L, R>::type,
      Lm-Rm, Lkg-Rkg, Ls-Rs, LA-RA, LK-RK, Lmol-Rmol, Lcd-Rcd>
      operator /(
          const ScalarNumber<L, Lm, Lkg, Ls, LA, LK, Lmol, Lcd>& lhs,
          const ScalarNumber<R, Rm, Rkg, Rs, RA, RK, Rmol, Rcd>& rhs
      )
{
    return ScalarNumber<typename BiggerDigitsType<L, R>::type,
           Lm-Rm, Lkg-Rkg, Ls-Rs, LA-RA, LK-RK, Lmol-Rmol, Lcd-Rcd>(
               lhs.X() / rhs.X());
}

/**
 * スカラーの除算.
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename T,
int m, int kg, int s, int A, int K, int mol, int cd
>
inline
const ScalarNumber<T, m, kg, s, A, K, mol, cd>
operator /(
    const ScalarNumber<T, m, kg, s, A, K, mol, cd>& lhs,
    const T& rhs
)
{
    return ScalarNumber<T, m, kg, s, A, K, mol, cd>(
               lhs.X() / rhs);
}

/**
 * スカラーの除算.
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename T,
         int m, int kg, int s, int A, int K, int mol, int cd
         >
inline
const ScalarNumber<T, -m, -kg, -s, -A, -K, -mol, -cd>
operator /(
    const T& lhs,
    const ScalarNumber<T, m, kg, s, A, K, mol, cd>& rhs
)
{
    return ScalarNumber<T, -m, -kg, -s, -A, -K, -mol, -cd>(
               lhs / rhs.X());
}

/**
 * スカラーの加算.
 * 次元が等しいスカラーのみ加算できる
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename L, typename R,
         int m, int kg, int s, int A, int K, int mol, int cd
         >
inline
const ScalarNumber<typename BiggerDigitsType<L, R>::type,
      m, kg, s, A, K, mol, cd>
      operator +(
          const ScalarNumber<L, m, kg, s, A, K, mol, cd>& lhs,
          const ScalarNumber<R, m, kg, s, A, K, mol, cd>& rhs
      )
{
    return ScalarNumber<typename BiggerDigitsType<L, R>::type,
           m, kg, s, A, K, mol, cd>(
               lhs.X() + rhs.X());
}

/**
 * スカラー単項演算子.
 * @param rhs 演算対象
 * @return 演算結果
 */
template <typename T,
int m, int kg, int s, int A, int K, int mol, int cd
>
inline
const ScalarNumber<T, m, kg, s, A, K, mol, cd>&
operator +(
    const ScalarNumber<T, m, kg, s, A, K, mol, cd>& rhs
)
{
    return rhs;
}

/**
 * スカラーの加算.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ加算できる
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename T>
inline
const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>
operator +(
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& lhs,
    const T& rhs
)
{
    return ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>(lhs.X() + rhs);
}

/**
 * スカラーの加算.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ加算できる
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename T>
inline
const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>
operator +(
    const T& lhs,
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& rhs
)
{
    return ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>(lhs + rhs.X());
}

/**
 * スカラーの減算.
 * 次元が等しいスカラーのみ減算できる
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename L, typename R,
         int m, int kg, int s, int A, int K, int mol, int cd
         >
inline
const ScalarNumber<typename BiggerDigitsType<L, R>::type,
      m, kg, s, A, K, mol, cd>
      operator -(
          const ScalarNumber<L, m, kg, s, A, K, mol, cd>& lhs,
          const ScalarNumber<R, m, kg, s, A, K, mol, cd>& rhs
      )
{
    return ScalarNumber<typename BiggerDigitsType<L, R>::type,
           m, kg, s, A, K, mol, cd>(
               lhs.X() - rhs.X());
}

/**
 * スカラー単項演算子.
 * @param rhs 演算対象
 * @return 演算結果
 */
template <typename T,
int m, int kg, int s, int A, int K, int mol, int cd
>
inline
const ScalarNumber<T, m, kg, s, A, K, mol, cd>
operator -(
    const ScalarNumber<T, m, kg, s, A, K, mol, cd>& rhs
)
{
    return ScalarNumber<T, m, kg, s, A, K, mol, cd>(-rhs.X());
}

/**
 * スカラーの減算.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ加算できる
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename T>
inline
const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>
operator -(
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& lhs,
    const T& rhs
)
{
    return ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>(lhs.X() - rhs);
}

/**
 * スカラーの減算.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ加算できる
 * @param lhs 演算の左側の値
 * @param rhs 演算の右側の値
 * @return 演算結果
 */
template <typename T>
inline
const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>
operator -(
    const T& lhs,
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& rhs
)
{
    return ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>(lhs - rhs.X());
}

/**
 * スカラーの大小比較.
 * 次元が等しいスカラーのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename L, typename R,
         int m, int kg, int s, int A, int K, int mol, int cd
         >
inline
bool operator ==(
    const ScalarNumber<L, m, kg, s, A, K, mol, cd>& lhs,
    const ScalarNumber<R, m, kg, s, A, K, mol, cd>& rhs
)
{
    typedef typename BiggerDigitsType<L, R>::type type;
    return static_cast<type>(lhs.X()) == static_cast<type>(rhs.X());
}

/**
 * スカラーの大小比較.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename T>
inline
bool operator ==(
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& lhs,
    const T& rhs
)
{
    return static_cast<T>(lhs.X()) == static_cast<T>(rhs);
}

/**
 * スカラーの大小比較.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename T>
inline
bool operator ==(
    const T& lhs,
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& rhs
)
{
    return static_cast<T>(lhs) == static_cast<T>(rhs.X());
}

/**
 * スカラーの大小比較.
 * 次元が等しいスカラーのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename L, typename R,
         int m, int kg, int s, int A, int K, int mol, int cd
         >
inline
bool operator >(
    const ScalarNumber<L, m, kg, s, A, K, mol, cd>& lhs,
    const ScalarNumber<R, m, kg, s, A, K, mol, cd>& rhs
)
{
    typedef typename BiggerDigitsType<L, R>::type type;
    return static_cast<type>(lhs.X()) > static_cast<type>(rhs.X());
}

/**
 * スカラーの大小比較.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename T>
inline
bool operator >(
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& lhs,
    const T& rhs
)
{
    return static_cast<T>(lhs.X()) > static_cast<T>(rhs);
}

/**
 * スカラーの大小比較.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename T>
inline
bool operator >(
    const T& lhs,
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& rhs
)
{
    return static_cast<T>(lhs) > static_cast<T>(rhs.X());
}

/**
 * スカラーの大小比較.
 * 次元が等しいスカラーのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename L, typename R,
         int m, int kg, int s, int A, int K, int mol, int cd
         >
inline
bool operator >=(
    const ScalarNumber<L, m, kg, s, A, K, mol, cd>& lhs,
    const ScalarNumber<R, m, kg, s, A, K, mol, cd>& rhs
)
{
    typedef typename BiggerDigitsType<L, R>::type type;
    return static_cast<type>(lhs.X()) >= static_cast<type>(rhs.X());
}

/**
 * スカラーの大小比較.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename T>
inline
bool operator >=(
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& lhs,
    const T& rhs
)
{
    return static_cast<T>(lhs.X()) >= static_cast<T>(rhs);
}

/**
 * スカラーの大小比較.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename T>
inline
bool operator >=(
    const T& lhs,
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& rhs
)
{
    return static_cast<T>(lhs) >= static_cast<T>(rhs.X());
}

/**
 * スカラーの大小比較.
 * 次元が等しいスカラーのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename L, typename R,
         int m, int kg, int s, int A, int K, int mol, int cd
         >
inline
bool operator <(
    const ScalarNumber<L, m, kg, s, A, K, mol, cd>& lhs,
    const ScalarNumber<R, m, kg, s, A, K, mol, cd>& rhs
)
{
    typedef typename BiggerDigitsType<L, R>::type type;
    return static_cast<type>(lhs.X()) < static_cast<type>(rhs.X());
}

/**
 * スカラーの大小比較.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename T>
inline
bool operator <(
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& lhs,
    const T& rhs
)
{
    return static_cast<T>(lhs.X()) < static_cast<T>(rhs);
}

/**
 * スカラーの大小比較.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename T>
inline
bool operator <(
    const T& lhs,
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& rhs
)
{
    return static_cast<T>(lhs) < static_cast<T>(rhs.X());
}

/**
 * スカラーの大小比較.
 * 次元が等しいスカラーのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename L, typename R,
         int m, int kg, int s, int A, int K, int mol, int cd
         >
inline
bool operator <=(
    const ScalarNumber<L, m, kg, s, A, K, mol, cd>& lhs,
    const ScalarNumber<R, m, kg, s, A, K, mol, cd>& rhs
)
{
    typedef typename BiggerDigitsType<L, R>::type type;
    return static_cast<type>(lhs.X()) <= static_cast<type>(rhs.X());
}

/**
 * スカラーの大小比較.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename T>
inline
bool operator <=(
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& lhs,
    const T& rhs
)
{
    return static_cast<T>(lhs.X()) <= static_cast<T>(rhs);
}

/**
 * スカラーの大小比較.
 * スカラー型と組み込み型は、スカラーが無次元数のときのみ比較できる
 * @param lhs 比較の左側の値
 * @param rhs 比較の右側の値
 * @return 比較結果
 */
template <typename T>
inline
bool operator <=(
    const T& lhs,
    const ScalarNumber<T, 0, 0, 0, 0, 0, 0, 0>& rhs
)
{
    return static_cast<T>(lhs) <= static_cast<T>(rhs.X());
}


using std::atan2;

template <typename L, typename R>
inline const AngleBase<typename BiggerDigitsType<L, R>::type>
atan2(const LengthBase<L>& y, const LengthBase<R>& x)
{
    return AngleBase<typename BiggerDigitsType<L, R>::type>(atan2(y.ToMeter(), x.ToMeter()));
}

using std::fabs;

template <typename T, int m, int kg, int s, int A, int K, int mol, int cd>
inline
const ScalarNumber<T, m, kg, s, A, K, mol, cd>
fabs(
    const ScalarNumber<T, m, kg, s, A, K, mol, cd>& x
)
{
    return ScalarNumber<T, m, kg, s, A, K, mol, cd>(fabs(x.X()));
}

using std::sqrt;

/**
 * 二乗根.
 * 二乗根を取った結果は、元の数値の各次元が2で割られた次元となるため、
 * 偶数ではない次元があった場合、結果がおかしな次元になってしまうので注意。
 */
template <typename T, int m, int kg, int s, int A, int K, int mol, int cd>
inline
const ScalarNumber<T, m/2, kg/2, s/2, A/2, K/2, mol/2, cd/2>
sqrt(
    const ScalarNumber<T, m, kg, s, A, K, mol, cd>& x
)
{
    return ScalarNumber<T, m/2, kg/2, s/2, A/2, K/2, mol/2, cd/2>(sqrt(x.X()));
}

}

#endif /* DIMPRO_OPERATORS_HPP_ */
