/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * IntegerOperator.hpp
 *
 *  Created on: 2011/01/31
 *      Author: HARK Development Team
 */

#ifndef INTEGEROPERATOR_HPP_
#define INTEGEROPERATOR_HPP_

/**
 * @file IntegerOperator.hpp
 * 安全な整数演算を提供する.
 */

#include <limits>
#include <stdexcept>

#include <iostream>

namespace NSSafeCalculation
{

template <typename T>
T OverflowAdd(const T& lhs, const T& rhs) throw (std::overflow_error)
{
    if ((lhs > 0 && rhs > 0 && lhs > std::numeric_limits<T>::max() - rhs) ||
            (lhs < 0 && rhs < 0 && lhs < std::numeric_limits<T>::min() - rhs))
    {
        throw std::overflow_error("addition overflow");
    }
    return lhs + rhs;
}

template <typename T>
T OverflowSub(const T& lhs, const T& rhs) throw (std::overflow_error)
{
    if ((lhs > 0 && rhs < 0 && lhs > std::numeric_limits<T>::max() + rhs) ||
            (lhs < 0 && rhs > 0 && lhs < std::numeric_limits<T>::min() + rhs))
    {
        throw std::overflow_error("subtraction overflow");
    }
    return lhs - rhs;
}

template <typename T>
T OverflowMul(const T& lhs, const T& rhs) throw (std::overflow_error)
{
    if ((lhs > 0 && rhs > 0 && lhs > std::numeric_limits<T>::max() / rhs) ||
            (lhs > 0 && rhs < 0 && lhs > std::numeric_limits<T>::min() / rhs) ||
            (lhs < 0 && rhs > 0 && lhs < std::numeric_limits<T>::min() / rhs) ||
            (lhs < 0 && rhs < 0 && lhs < std::numeric_limits<T>::max() / rhs))
    {
        throw std::overflow_error("multiplication overflow");
    }
    return lhs * rhs;
}

template <typename T>
T OverflowDiv(const T& lhs, const T& rhs) throw (std::overflow_error, std::logic_error)
{
    if ((lhs > 0 && rhs > 0 && lhs > std::numeric_limits<T>::max() * rhs) ||
            (lhs > 0 && rhs < 0 && lhs > std::numeric_limits<T>::min() * rhs) ||
            (lhs < 0 && rhs > 0 && lhs < std::numeric_limits<T>::min() * rhs) ||
            (lhs < 0 && rhs < 0 && lhs < std::numeric_limits<T>::max() * rhs))
    {
        throw std::overflow_error("division overflow");
    }
    else if (rhs == 0)
    {
        throw std::logic_error("division by zero");
    }
    return lhs / rhs;
}

template <typename T>
T NoOverflowAdd(const T& lhs, const T& rhs) throw ()
{
    if (lhs > 0 && rhs > 0 && lhs > std::numeric_limits<T>::max() - rhs)
    {
        return std::numeric_limits<T>::max();
    }
    else if (lhs < 0 && rhs < 0 && lhs < std::numeric_limits<T>::min() - rhs)
    {
        return std::numeric_limits<T>::min();
    }
    return lhs + rhs;
}

template <typename T>
T NoOverflowSub(const T& lhs, const T& rhs) throw ()
{
    if (lhs > 0 && rhs < 0 && lhs > std::numeric_limits<T>::max() + rhs)
    {
        return std::numeric_limits<T>::max();
    }
    else if (lhs < 0 && rhs > 0 && lhs < std::numeric_limits<T>::min() + rhs)
    {
        return std::numeric_limits<T>::min();
    }
    return lhs - rhs;
}

template <typename T>
T NoOverflowMul(const T& lhs, const T& rhs) throw ()
{
    if ((lhs > 0 && rhs > 0 && lhs > std::numeric_limits<T>::max() / rhs) ||
            (lhs < 0 && rhs < 0 && lhs < std::numeric_limits<T>::max() / rhs))
    {
        return std::numeric_limits<T>::max();
    }
    else if ((lhs > 0 && rhs < 0 && lhs > std::numeric_limits<T>::min() / rhs) ||
             (lhs < 0 && rhs > 0 && lhs < std::numeric_limits<T>::min() / rhs))
    {
        return std::numeric_limits<T>::min();
    }
    return lhs * rhs;
}

template <typename T>
T NoOverflowDiv(const T& lhs, const T& rhs) throw (std::logic_error)
{
    if ((lhs > 0 && rhs > 0 && lhs > std::numeric_limits<T>::max() * rhs) ||
            (lhs < 0 && rhs < 0 && lhs < std::numeric_limits<T>::max() * rhs))
    {
        return std::numeric_limits<T>::max();
    }
    else if ((lhs > 0 && rhs < 0 && lhs > std::numeric_limits<T>::min() * rhs) ||
             (lhs < 0 && rhs > 0 && lhs < std::numeric_limits<T>::min() * rhs))
    {
        return std::numeric_limits<T>::min();
    }
    else if (rhs == 0)
    {
        throw std::logic_error("division by zero");
    }
    return lhs / rhs;
}

namespace impl
{

template <typename T, typename U, bool TIsSigned, bool UIsSigned, bool TIsBigger>
struct overflow_cast_impl
{
    // ビット数が増える方向のキャスト
    static T Cast(const U& val) throw () {
        return static_cast<T>(val);
    }
};

template <typename T, typename U>
struct overflow_cast_impl<T, U, false, false, false>
{
    // ビット数が減る方向のキャスト
    // T,Uは符号なし
    static T Cast(const U& val) throw (std::overflow_error)
    {
        if (val <= std::numeric_limits<T>::max())
        {
            return static_cast<T>(val);
        }
        throw std::overflow_error("cast overflow");
    }
};

template <typename T, typename U>
struct overflow_cast_impl<T, U, false, true, false>
{
    // ビット数が減る方向のキャスト
    // Tは符号なし, Uは符号付き
    static T Cast(const U& val) throw (std::overflow_error)
    {
        if (val >= 0 && ((std::numeric_limits<T>::digits == std::numeric_limits<U>::digits) || (val <= static_cast<U>(std::numeric_limits<T>::max()))))
        {
            return static_cast<T>(val);
        }
        throw std::overflow_error("cast overflow");
    }
};

template <typename T, typename U>
struct overflow_cast_impl<T, U, true, false, false>
{
    // ビット数が減る方向のキャスト
    // Tは符号付き, Uは符号なし
    static T Cast(const U& val) throw (std::overflow_error)
    {
        if (val <= static_cast<U>(std::numeric_limits<T>::max()))
        {
            return static_cast<T>(val);
        }
        throw std::overflow_error("cast overflow");
    }
};

template <typename T, typename U>
struct overflow_cast_impl<T, U, true, true, false>
{
    // ビット数が減る方向のキャスト
    // T,Uは符号付き
    static T Cast(const U& val) throw (std::overflow_error)
    {
        if (std::numeric_limits<T>::min() <= val && val <= std::numeric_limits<T>::max())
        {
            return static_cast<T>(val);
        }
        throw std::overflow_error("cast overflow");
    }
};

template <typename T, typename U, bool TIsSigned, bool UIsSigned, bool TIsBigger>
struct no_overflow_cast_impl
{
    // ビット数が増える方向のキャスト
    static T Cast(const U& val) throw () {
        return static_cast<T>(val);
    }
};

template <typename T, typename U>
struct no_overflow_cast_impl<T, U, false, false, false>
{
    // ビット数が減る方向のキャスト
    // T,Uは符号なし
    static T Cast(const U& val) throw ()
    {
        if (val <= std::numeric_limits<T>::max())
        {
            return static_cast<T>(val);
        }
        return std::numeric_limits<T>::max();
    }
};

template <typename T, typename U>
struct no_overflow_cast_impl<T, U, false, true, false>
{
    // ビット数が減る方向のキャスト
    // Tは符号なし, Uは符号付き
    static T Cast(const U& val) throw ()
    {
        if (val >= 0 && ((std::numeric_limits<T>::digits == std::numeric_limits<U>::digits) || (val <= static_cast<U>(std::numeric_limits<T>::max()))))
        {
            return static_cast<T>(val);
        }
        else if (val >= 0)
        {
            return std::numeric_limits<T>::max();
        }
        return 0;
    }
};

template <typename T, typename U>
struct no_overflow_cast_impl<T, U, true, false, false>
{
    // ビット数が減る方向のキャスト
    // Tは符号付き, Uは符号なし
    static T Cast(const U& val) throw ()
    {
        if (val <= static_cast<U>(std::numeric_limits<T>::max()))
        {
            return static_cast<T>(val);
        }
        return std::numeric_limits<T>::max();
    }
};

template <typename T, typename U>
struct no_overflow_cast_impl<T, U, true, true, false>
{
    // ビット数が減る方向のキャスト
    // T,Uは符号付き
    static T Cast(const U& val) throw ()
    {
        if (std::numeric_limits<T>::min() <= val && val <= std::numeric_limits<T>::max())
        {
            return static_cast<T>(val);
        }
        else if (val < static_cast<U>(std::numeric_limits<T>::min()))
        {
            return std::numeric_limits<T>::min();
        }
        return std::numeric_limits<T>::max();
    }
};

}

/**
 * オーバーフローすると例外を発生させるstatic_cast.
 */
template <typename T>
struct overflow_static_cast
{
    T val;

    template <typename U>
    overflow_static_cast(const U& val) throw (std::overflow_error)
        : val(impl::overflow_cast_impl<T, U, std::numeric_limits<T>::is_signed, std::numeric_limits<U>::is_signed,
              (std::numeric_limits<T>::digits > std::numeric_limits<U>::digits)
              >::Cast(val))
    {
    }

    operator T() const {
        return val;
    }
};

/**
 * オーバーフローするとキャスト先の型の最大または最小値になるstatic_cast.
 */
template <typename T>
struct no_overflow_static_cast
{
    T val;

    template <typename U>
    no_overflow_static_cast(const U& val) throw ()
        : val(impl::no_overflow_cast_impl<T, U, std::numeric_limits<T>::is_signed, std::numeric_limits<U>::is_signed,
              (std::numeric_limits<T>::digits > std::numeric_limits<U>::digits)
              >::Cast(val))
    {
    }

    operator T() const {
        return val;
    }
};

}

#endif /* INTEGEROPERATOR_HPP_ */
