/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * EventHistoryManager.cpp
 *
 *  Created on: 2011/01/24
 *      Author: HARK Development Team
 */

#include <stdexcept>
#include <DimensionPromise/dimpro.hpp>
#include <EventHistoryManager.hpp>
#include <IntegerOperator.hpp>
#include <Timeval.hpp>
#include <DirectionConverter.hpp>

using namespace std;
using namespace boost;
using namespace NSDimensionPromise;
using namespace NSSafeCalculation;

namespace NSHarkBinaural
//namespace // 無名ネームスペース
{

template <typename Iterator>
Iterator SearchEID(Iterator begin, Iterator end, EventID eid)
{
    Iterator it = begin;
    for (; it != end; ++it)
    {
        if (it->ID() == eid)
        {
            break;
        }
    }
    return it;
}

/**
 * イベントヒストリーを探す.
 * @param eventHistories イベントヒストリーのリスト
 * @param eid 探索するキー
 * @param begin eventHistories内での探索開始位置
 * @return begin以降で始めて見つかった位置
 */
std::list<EventHistory>::iterator
Search(std::list<EventHistory>& eventHistories, EventID eid,
       std::list<EventHistory>::iterator begin)
{
    list<EventHistory>::iterator it = begin;
    for (; it != eventHistories.end(); ++it)
    {
        if (it->ID() == eid)
        {
            break;
        }
    }
    return it;
}

/**
 * イベントヒストリー更新情報を探す.
 * @param updateInfoList イベントヒストリー更新情報のリスト
 * @param eid 探索するキー
 * @param begin updateInfoList内での探索開始位置
 * @return begin以降で始めて見つかった位置
 */
std::vector<EventHistoryManager::UpdateInfo>::iterator
Search(std::vector<EventHistoryManager::UpdateInfo>& updateInfoList, EventID eid,
       std::vector<EventHistoryManager::UpdateInfo>::iterator begin)
{
    vector<EventHistoryManager::UpdateInfo>::iterator it = begin;
    for (; it != updateInfoList.end(); ++it)
    {
        if (tuples::get<0>(*it) == eid)
        {
            break;
        }
    }
    return it;
}

template <typename T>
T TimevalToMilliseconds(const timeval& tim, bool useException = false)
{
    /*
    if (tim.tv_sec >= numeric_limits<T>::max() / 1000 - tim.tv_usec / 1000000)
    {
      throw out_of_range("tim is too big");
    }
    else if (tim.tv_sec <= numeric_limits<T>::min() / 1000 - tim.tv_usec / 1000000)
    {
      throw out_of_range("tim is too small");
    }
    return static_cast<T>(tim.tv_sec * 1000 + tim.tv_usec / 1000);
    */

    // tim.tv_sec * 1000 + tim.tv_usec / 1000
    if (useException)
    {
        return OverflowAdd<T>(OverflowMul<T>(tim.tv_sec, 1000), OverflowDiv<T>(tim.tv_usec, 1000));
    }
    else
    {
        return NoOverflowAdd<T>(NoOverflowMul<T>(tim.tv_sec, 1000), NoOverflowDiv<T>(tim.tv_usec, 1000));
    }
}

}

namespace NSHarkBinaural
{

EventHistoryManager::EventHistoryManager()
    : logger(Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME)),
      eventHistories(),
      nearDirectionThreshold(Angle::FromDegree(6)),
      validPausePeriod(),
      headid(0),
      eventUpdateRatioThreshold(0.0)
{
    this->validPausePeriod.tv_sec = 1;
    this->validPausePeriod.tv_usec = 0;
}

void EventHistoryManager::CheckSoundHistory(HarmonicSoundVectorPtr sounds, const timeval& currentTime)
{
    // checkSoundHistory関数のhark-binaural版
    if (!sounds)
    {
        throw invalid_argument("sounds is null");
    }
    vector<UpdateInfo> updateInfoList;

    for (HarmonicSoundVector::const_iterator it = sounds->begin(); it != sounds->end(); ++it)
    {
        const HarmonicSoundPtr ptr = *it;
        Band fundamental = ptr->bands->at(0).band;
        Angle direction = Angle::FromDegree(gDirConv.ToDegree(ptr->directions->at(0).first));
        EventID eid = FindNearestEvent(fundamental, direction, currentTime);
        updateInfoList.push_back(make_tuple(eid, fundamental, direction));
    }
    Update(updateInfoList, currentTime);

    for (size_t i = 0; i < sounds->size(); ++i)
    {
        sounds->operator [](i)->eventID = tuples::get<0>(updateInfoList.at(i));
    }
}

void EventHistoryManager::CheckSoundHistory(boost::shared_ptr<vector<Angle> > directions, const timeval& currentTime)
{
    // checkSoundHistory関数のhark-binaural版
    if (!directions)
    {
        throw invalid_argument("directions is null");
    }
    vector<UpdateInfo> updateInfoList;

    for (vector<Angle>::const_iterator it = directions->begin(); it != directions->end(); ++it)
    {
        Band fundamental = Band();
        Angle direction = *it;
        EventID eid = FindNearestEvent(fundamental, direction, currentTime);
        updateInfoList.push_back(make_tuple(eid, fundamental, direction));
    }
    Update(updateInfoList, currentTime);
}

void EventHistoryManager::Update(std::vector<UpdateInfo>& updateInfoList, const timeval& currentTime)
{
    for (list<EventHistory>::iterator eh = this->eventHistories.begin(); eh != this->eventHistories.end(); ++eh)
    {
        vector<UpdateInfo>::iterator ui = Search(updateInfoList, eh->ID(), updateInfoList.begin());
        if (eh->ID().IsConnected() && ui != updateInfoList.end())
        {
            if (!IsNear(eh->Direction(), tuples::get<2>(*ui)))
            {
                LogSevere(logger, "event history direction is not near of update info direction");
                throw logic_error("eh is not near ui");
            }
            eh->Update(currentTime, tuples::get<1>(*ui), tuples::get<2>(*ui));
        }
        else // コネクトするものがない
        {
            eh->Update(currentTime);
        }
    }
    EraseOutOfDateEvents();

    MakeNewEvents(updateInfoList, currentTime);
}

EventID EventHistoryManager::FindNearestEvent(Band fundamental, NSDimensionPromise::Angle direction, const timeval& currentTime)
{
    int fdiff;
    Angle dirdiff;
    list<EventHistory>::iterator minItem = this->eventHistories.end();

    timeval sub;
    int minval = 1000; // 評価値の最小値
    int val; // 評価値

    for (list<EventHistory>::iterator it = this->eventHistories.begin(); it != this->eventHistories.end(); ++it)
    {
        fdiff = abs(it->Fundamental().Index() - fundamental.Index()); // 周波数の差
        dirdiff = fabs(it->Direction() - direction); // 角度の差
        if (fdiff < 3 && IsNear(it->Direction(), direction))
        {
            // 候補発見
            if (it->Status() == EventStatus::PAUSE)
            {
                //timersub(&currentTime, &it->pauseTime, &sub);
                //val = sub.tv_usec / 1000 + sub.tv_sec * 1000 + fdiff;
                val = NoOverflowAdd<int>(TimevalToMilliseconds<int>(currentTime - it->PauseTime()), fdiff);
            }
            else
            {
                val = fdiff;
            }
            if (val < minval)
            {
                minval = val;
                minItem = it;
                if (minval == 0)
                {
                    break;
                }
            }
        }
    }

    if (minItem != this->eventHistories.end())
    {
        return minItem->ID();
    }
    return EventID();
}

bool EventHistoryManager::IsNear(NSDimensionPromise::Angle dir1, NSDimensionPromise::Angle dir2)
{
    return fabs( dir1 - dir2 ) <= nearDirectionThreshold;
}

void EventHistoryManager::EraseOutOfDateEvents()
{
    for (list<EventHistory>::iterator it = this->eventHistories.begin(); it != this->eventHistories.end(); )
    {
        if (it->IsOutOfDate())
        {
            LogFinest(logger, format("erase event (eid = %d)") % static_cast<int>(it->ID()));
            it = this->eventHistories.erase(it);
            continue;
        }
        ++it;
    }
}

void EventHistoryManager::MakeNewEvents(std::vector<UpdateInfo>& updateInfoList, const timeval& currentTime)
{
    for (vector<UpdateInfo>::iterator it = updateInfoList.begin(); it != updateInfoList.end(); ++it)
    {
        if (!tuples::get<0>(*it).IsConnected())
        {
            tuples::get<0>(*it) = MakeNewEvent(currentTime, tuples::get<1>(*it), tuples::get<2>(*it));
        }
    }
}

EventID EventHistoryManager::MakeNewEvent(const timeval& currentTime, Band fundamental, NSDimensionPromise::Angle direction)
{
    EventHistory eh(NextEID(), currentTime, this->validPausePeriod, fundamental, direction, this->eventUpdateRatioThreshold);
    this->eventHistories.push_back(eh);
    return eh.ID();
}

EventID EventHistoryManager::NextEID()
{
    if (0 <= this->headid && this->headid < numeric_limits<int>::max())
    {
        this->headid++;
        return EventID(this->headid);
    }
    throw overflow_error("can't create new eid");
}

}
